import abc
import functools
from typing import Callable, TypeVar, Any, Generic, Type, Tuple, Union
from asyncio import iscoroutinefunction

from amino.tc.optional import Optional
from amino import Maybe, may, Either, Just, Left, I, List, Nothing, _, __, Id
from amino.task import TaskException
from amino.util.exception import format_exception
from amino.state import StateT
from amino.util.string import ToStr

from ribosome.machine.message_base import (_message_attr, _machine_attr, Message, default_prio, _prio_attr,
                                           fallback_prio, override_prio, _dyn_attr)

from ribosome.record import Record, any_field, list_field, field, bool_field, optional_field
from ribosome.logging import Logging
from ribosome.machine.interface import MachineI
from ribosome.data import Data
from ribosome.machine.transitions import Transitions
from ribosome.machine.trans import TransAction
from ribosome.machine.messages import Error, Debug, Coroutine

A = TypeVar('A')


class Failure:

    def __init__(self, message: str) -> None:
        self.message = message

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.message)


class Fatal(Failure):
    pass


class NothingToDo(Failure):
    pass


def _to_error(data):
    return (
        data if isinstance(data, Message)
        else Error(data.message) if isinstance(data, Fatal)
        else Debug(data.message) if isinstance(data, NothingToDo)
        else Error(data) if isinstance(data, Exception)
        else Debug(str(data))
    )


def _recover_error(handler: Any, result: Any) -> Maybe[Any]:
    if not Optional.exists(type(result)):
        err = 'in {}: result has no Optional: {}'
        return Just(Error(err.format(handler, result)))
    elif isinstance(result, Either):
        return Just(result.right_or_map(_to_error))
    else:
        return result


def _task_result(result):
    return result.cata(Left, I)

D = TypeVar('D', bound=Data)
M = TypeVar('M', bound=Message)
R = TypeVar('R')
T = TypeVar('T', bound=Transitions)
Msg = TypeVar('Msg', bound=Message)
DynResult = Union[Message, List[Message]]
DynTrans = Union[DynResult, Maybe[DynResult], Either[str, DynResult], StateT[Id, Data, DynResult]]


class Handler(Generic[M, D, R], Logging, ToStr):

    def __init__(self, machine: MachineI, name: str, fun: Callable[[MachineI, D, M], R], message: Type[M], prio: float,
                 dyn: bool) -> None:
        self.machine = machine
        self.name = name
        self.message = message
        self.fun = fun
        self.prio = prio
        self.dyn = dyn

    @staticmethod
    def attrs(fun) -> Tuple[Type[M], float, bool]:
        msg = getattr(fun, _message_attr)
        prio = getattr(fun, _prio_attr, default_prio)
        dyn = getattr(fun, _dyn_attr)
        return msg, prio, dyn

    @staticmethod
    def create(machine: MachineI, name: str, fun: Callable[[MachineI, D, M], R]) -> 'Handler[M, D, R]':
        msg, prio, dyn = Handler.attrs(fun)
        tpe = CoroHandler if iscoroutinefunction(fun) else DynHandler  # if dyn else AlgHandler
        return tpe(machine, name, fun, msg, prio, dyn)

    @abc.abstractmethod
    def execute(self, data: D, msg: Msg) -> R:
        ...

    @abc.abstractmethod
    def run(self, data: D, msg: Msg) -> R:
        ...

    def _arg_desc(self) -> List[str]:
        return List(self.name, str(self.message), str(self.fun))


class DynHandler(Generic[M, D], Handler[M, D, DynTrans]):

    def run(self, data, msg) -> DynTrans:
        return _recover_error(self, self.execute(data, msg))

    def execute(self, data, msg) -> DynTrans:
        return self.fun(self.machine, data, msg)


class WrappedHandler(Generic[M, D, T], DynHandler[M, D]):

    def __init__(self, trans_tpe: Type[T], machine: MachineI, name: str, fun: Callable[[MachineI, D, M], DynTrans],
                 message: Type[M], prio: float, dyn: bool) -> None:
        super().__init__(machine, name, lambda mach, d, msg: fun(trans_tpe(mach, d, msg)), message, prio, dyn)

    @staticmethod
    def create(machine: MachineI, name: str, fun: Callable[[MachineI, D, M], R], tpe: Type[T]
               ) -> 'Handler[M, D, DynTrans]':
        msg, prio, dyn = Handler.attrs(fun)
        return WrappedHandler(tpe, machine, name, fun, msg, prio, dyn)


class AlgHandler(Generic[M, D], Handler[M, D, TransAction]):

    def run(self, data: D, msg: Msg) -> TransAction:
        ...

    def execute(self, data: D, msg: Msg) -> TransAction:
        return self.fun(self.machine, data, msg)


class CoroHandler(Handler):

    def run(self, data, msg) -> Maybe[Coroutine]:
        return Maybe(Coroutine(self.fun(data, msg)))

    def execute(self, data, msg) -> Maybe[Coroutine]:
        ...


class CoroExecutionHandler(DynHandler):
    pass


class TransitionResult(Record):
    machine = optional_field(MachineI)
    data = any_field()
    resend = list_field()
    handled = bool_field(True)
    failure = bool_field(False)
    error = optional_field()

    @staticmethod
    def empty(data, **kw):
        return StrictTransitionResult(data=data, **kw)

    @staticmethod
    def failed(data, error, **kw):
        return TransitionResult.unhandled(data, failure=True, error=Just(error), **kw)

    @staticmethod
    def unhandled(data, **kw):
        return StrictTransitionResult(data=data, handled=False, **kw)

    @property
    def _str_extra(self):
        return List(self.handled) + self.error.to_list + self.resend

    def fold(self, f):
        return self

    def accum(self, other: 'TransitionResult'):
        if isinstance(other, CoroTransitionResult):
            return self.accum(StrictTransitionResult(
                data=other.data,
                pub=other.pub,
                handled=other.handled or self.handled
            ))
        else:
            return other.set(
                pub=self.pub + other.pub,
                handled=other.handled or self.handled
            )

    @property
    def error_message(self):
        def format(err):
            return str(err.cause if isinstance(err, TaskException) else err)
        return self.error / format | 'unknown error'

    @property
    def exception(self) -> Maybe[Exception]:
        def analyze(e: Exception) -> Exception:
            return (
                e.__cause__
                if isinstance(e, TransitionFailed) and e.__cause__ is not None else
                e.cause
                if isinstance(e, TaskException) else
                e
            )
        return self.error // (lambda err: Just(analyze(err)) if isinstance(err, Exception) else Nothing)

    @property
    def exception_fmt(self) -> Maybe[str]:
        return (self.exception / format_exception) / __.cons('Exception:') / _.join_lines


class StrictTransitionResult(TransitionResult):
    pub = list_field()

    def fold(self, f):
        return self.resend.fold_left(self)(f)

    @property
    def _str_extra(self) -> List[Any]:
        return super()._str_extra + self.pub


class CoroTransitionResult(TransitionResult, Logging):
    coro = field(Coroutine)

    @property
    def pub(self):
        return [self.coro]


def handle(msg: type, prio=default_prio):
    def add_handler(func: Callable[..., Maybe[A]]):
        setattr(func, _machine_attr, True)
        setattr(func, _message_attr, msg)
        setattr(func, _prio_attr, prio)
        setattr(func, _dyn_attr, True)
        return func
    return add_handler


def may_handle(msg: type, prio=default_prio):
    def may_wrap(func: Callable[..., Maybe[A]]):
        return handle(msg, prio=prio)(may(func))
    return may_wrap


def fallback(msg: type, prio=fallback_prio):
    return handle(msg, prio=prio)


def may_fallback(msg: type, prio=fallback_prio):
    return may_handle(msg, prio=prio)


def override(msg: type, prio=override_prio):
    return handle(msg, prio=prio)


def may_override(msg: type, prio=override_prio):
    return may_handle(msg, prio=prio)


def either_msg(e: Either):
    return e.right_or_map(Error)


def either_handle(msg: type):
    def decorator(f):
        @functools.wraps(f)
        def either_wrap(*args, **kwargs):
            return Just(either_msg(f(*args, **kwargs)))
        return handle(msg)(either_wrap)
    return decorator


class MachineError(RuntimeError):
    pass


class TransitionFailed(MachineError):
    pass

__all__ = ('Handler', 'CoroHandler', 'TransitionResult', 'StrictTransitionResult', 'CoroTransitionResult', 'handle',
           'may_handle', 'either_msg', 'either_handle', 'MachineError', 'TransitionFailed')



import inspect
from types import FunctionType

from amino import Map, Maybe, Lists, List, Path
from amino.util.string import camelcaseify

from ribosome.nvim import NvimFacade
from ribosome.logging import ribo_log


class HandlerSpec:

    def __init__(self, tpe: str, sync: int, name: str, opts: dict) -> None:
        self.tpe = tpe
        self.sync = int(sync) if isinstance(sync, bool) else sync
        self.name = name
        self.opts = opts

    @staticmethod
    def cons(data: Map) -> Maybe['HandlerSpec']:
        return data.get_all('type', 'sync', 'name', 'opts').map4(HandlerSpec)


def handler(method_name: str, fun: FunctionType) -> Maybe[HandlerSpec]:
    return Maybe(getattr(fun, '_nvim_rpc_spec', None)) / Map // HandlerSpec.cons


def register_handler_args(host: str, spec: HandlerSpec, plugin_file: Path) -> List[str]:
    pat = spec.opts.get('pattern', '*')
    suf = f':{pat}' if type == 'autocmd' else ''
    rpc_method = f'{plugin_file}:{spec.tpe}:{spec.name}{suf}'
    fun_prefix = camelcaseify(spec.tpe)
    return List(f'remote#define#{fun_prefix}OnHost', host, rpc_method, spec.sync, spec.name, spec.opts)


def setup_rpc(vim: NvimFacade, name: str, plugin_class: type) -> None:
    plugin_file = inspect.getfile(plugin_class)
    ribo_log.test('-------------')
    ribo_log.test(plugin_file)
    if not vim.command_exists(f'{camelcaseify(name)}Start'):
        rp_handlers = Lists.wrap(inspect.getmembers(plugin_class)).flat_map2(handler)
        for spec in rp_handlers:
            ribo_log.test(spec)
            args = register_handler_args(name, spec, plugin_file)
            ribo_log.test(args)
            r = vim.call(*args)
            ribo_log.test(r)

__all__ = ('setup_rpc',)



import uuid
import json
from typing import Callable, Union, Pattern, Tuple, TypeVar

import pyrsistent
from typing import Any

from lenses import Lens, lens

from toolz import merge

from amino import (List, Empty, Boolean, _, Map, Left, L, __, Either, Try, Maybe, Just, Path, I, Regex)
from amino.lazy import LazyMeta, Lazy, lazy
from amino.lazy_list import LazyList
from amino.tc.optional import Optional
from amino.tc.foldable import Foldable

from ribosome.logging import Logging

A = TypeVar('A')


def any_field(**kw: Any) -> pyrsistent.field:
    return pyrsistent.field(mandatory=True, **kw)


def field(tpe: type, **kw: Any) -> pyrsistent.field:
    return any_field(type=tpe, **kw)


def _foldable_type_field_inv(eff: Any, tpe: type) -> Callable[[Any], Tuple[bool, str]]:
    def inv(a: Any) -> Tuple[bool, str]:
        atpe = type(a)
        if tpe is None:
            return True, ''
        elif not Foldable.exists_instance(atpe):
            return False, '{} does not have a Foldable'.format(atpe)
        else:
            bad = a.find(lambda b: not isinstance(b, tpe))
            bad_tpe = bad | ''
            err = 'must be {}[{}], found {}'.format(eff, tpe.__name__, bad_tpe)
            return not bad.present, err
    return inv


def list_field(tpe=None, initial=List(), **kw: Any) -> pyrsistent.field:
    return field(List, initial=initial, factory=List.wrap, invariant=_foldable_type_field_inv('List', tpe), **kw)


def lazy_list_field(**kw: Any) -> pyrsistent.field:
    return field(LazyList, initial=LazyList([]), factory=LazyList, **kw)


def dfield(default: Any, **kw: Any) -> pyrsistent.field:
    return field(type(default), initial=default, **kw)


def maybe_factory(fact, tpe) -> Callable[[Maybe[A]], bool]:
    def f(val: Maybe[A]) -> bool:
        return val / (lambda v: v if isinstance(v, tpe) else fact(v))
    return f


def maybe_field(tpe=None, factory=None, initial=Empty(), **kw: Any) -> pyrsistent.field:
    fact = I if factory is None or tpe is None else maybe_factory(factory, tpe)
    return field(Maybe, initial=initial,
                 invariant=_foldable_type_field_inv('Maybe', tpe),
                 factory=fact, **kw)


def either_field(rtpe, ltpe=str, initial=Left('pristine'), **kw: Any) -> pyrsistent.field:
    err = 'must be Either[{}, {}]'.format(ltpe, rtpe)
    check = lambda t, a: (isinstance(a, t), err)
    inv = __.cata(L(check)(ltpe, _), L(check)(rtpe, _))
    return field(Either, initial=initial, invariant=inv, **kw)


class OptionalInvariant:

    def __init__(self, tpe) -> None:
        self.tpe = tpe

    def __call__(self, a):
        atpe = type(a)
        return (
            _foldable_type_field_inv('Optional', self.tpe)(a)
            if Optional.exists_instance(atpe) else
            (False, '{} has no Optional'.format(atpe))
        )


class OptionalFactory:

    def __init__(self, factory, tpe) -> None:
        self.factory = factory
        self.tpe = tpe

    @property
    def fact(self):
        return maybe_factory(self.factory, self.tpe)

    def __call__(self, a):
        return a if self.factory is None or self.tpe is None else self.fact(a)


def optional_field(tpe=None, factory=None, initial=Empty(), **kw: Any) -> pyrsistent.field:
    return any_field(initial=initial, factory=OptionalFactory(factory, tpe),
                     invariant=OptionalInvariant(tpe), **kw)


def bool_field(initial=False, **kw: Any) -> pyrsistent.field:
    return field(Boolean, factory=Boolean.wrap, initial=Boolean(initial), **kw)


def map_field(**kw: Any) -> pyrsistent.field:
    return dfield(Map(), **kw)


def uuid_field() -> pyrsistent.field:
    return field(uuid.UUID, initial=lambda: uuid.uuid4())


def int_field(**kw: Any) -> pyrsistent.field:
    return field(int, **kw)


def float_field(**kw: Any) -> pyrsistent.field:
    return field(float, **kw)


def str_field(**kw: Any) -> pyrsistent.field:
    return field(str, **kw)


def _re_fact(rex: Union[str, Regex, Pattern]) -> Regex:
    return rex if isinstance(rex, Regex) else Regex.cons(rex)


def re_field(**kw: Any) -> pyrsistent.field:
    return field(Regex, factory=_re_fact, **kw)


class FieldMutator(object):

    def __init__(self, name: str, target: 'Record') -> None:
        self.name = name
        self.target = target


class FieldSetter(FieldMutator):

    def __call__(self, value):
        return self.target.set(**{self.name: value})


class FieldModifier(FieldMutator):

    def __call__(self, f):
        return self.target.set(
            **{self.name: f(getattr(self.target, self.name))})


class FieldAppender(FieldMutator):

    def __call__(self, value):
        return self.target.mod(self.name, _ + value)


class FieldAppender1(FieldMutator):

    def __call__(self, value):
        return self.target.mod(self.name, _ + List(value))


class FieldProxy(FieldMutator):

    def __init__(self, target: 'Record', tpe: type) -> None:
        self.target = target
        self.tpe = tpe

    def __getattr__(self, name):
        return self(name)

    def __call__(self, name):
        return self.tpe(name, self.target)


class RecordMeta(LazyMeta, pyrsistent.PClassMeta):

    @property
    def mandatory_fields(self):
        mand = lambda f: f.mandatory and f.initial.__class__ is object
        return Map(self._pclass_fields).valfilter(mand)

    @property
    def _field_map(self):
        return Map(self._pclass_fields)

    @property
    def _field_data(self):
        return self._field_map.valmap(_.type)

    @property
    def _field_names(self):
        return List.wrap(self._pclass_fields.keys())

    @property
    def _field_names_no_uuid(self):
        return self._field_names - 'uuid'


class Record(pyrsistent.PClass, Lazy, Logging, metaclass=RecordMeta):

    @classmethod
    def args_from_opt(cls, opt: Map):
        def may_arg(name, field, value):
            ctor = field.factory
            return name, ctor(Just(value))
        def list_arg(name, field, value):
            return name, value
        def regular_arg(name, field, value):
            return name, value
        def arg(name, field):
            cb = (
                may_arg if Maybe in field.type else
                list_arg if List in field.type else
                may_arg if isinstance(field.factory, OptionalFactory) else
                regular_arg
            )
            return opt.get(name) / L(cb)(name, field, _)
        return Map(cls._field_map.map2(arg).join)

    @classmethod
    def from_opt(cls, opt: Map):
        return cls(**cls.args_from_opt(opt))

    @classmethod
    def from_attr(cls, arg):
        return lambda a, **kw: cls(**{arg: a}, **kw)

    def set_from_opt(self, opt: Map):
        return self.set(**self.args_from_opt(opt))

    def update_from_opt(self, opt: Map):
        return self.set(**self.args_from_opt(opt))

    @property
    def _name(self):
        return self.__class__.__name__

    @property
    def _str_name(self):
        return self._name

    @property
    def _str_extra(self) -> List:
        return List()

    @property
    def _str_extra_named(self) -> Map:
        return Map()

    def __str__(self):
        extra_named = self._str_extra_named.map2('{}={}'.format)
        return '{}({})'.format(self._str_name,
                               (self._str_extra + extra_named).join_comma)

    @lazy
    def setter(self):
        return FieldProxy(self, FieldSetter)

    def _lens_setattr(self, name, value):
        return self.setter(name)(value)

    @lazy
    def append(self):
        return FieldProxy(self, FieldAppender)

    @lazy
    def append1(self):
        return FieldProxy(self, FieldAppender1)

    @lazy
    def modder(self):
        return FieldProxy(self, FieldModifier)

    def mod(self, name: str, modder):
        par = {name: modder(getattr(self, name))}
        return self.set(**par)

    @property
    def _fields_no_uuid(self):
        return type(self)._field_names_no_uuid / (lambda a: getattr(self, a))

    def __eq__(self, other):
        return (
            type(self) == type(other) and
            type(self)._field_names_no_uuid
            .forall(lambda a: getattr(self, a) == getattr(other, a))
        )

    def __hash__(self):
        return sum(self._fields_no_uuid / hash)

    def attr_lens(self, attr: Callable[..., List], sub: Callable[..., Lens]):
        return attr(self).find_lens(sub) / attr(lens()).add_lens

    def attr_lens_pred(self, attr: Callable[..., List],
                       pred: Callable[..., bool]):
        return attr(self).find_lens_pred(pred) / attr(lens()).add_lens

    @property
    def __path__(self):
        cls = self.__class__
        return '{}.{}'.format(cls.__module__, cls.__name__)

    def to_map(self):
        return Map(type(self)._field_names_no_uuid.apzip(L(getattr)(self, _)))

    @property
    def json(self):
        return merge(self.to_map(), dict(__type__=self.__path__))


def _decode_json_obj(obj):
    m = Map(obj)
    return (
        m.get('__type__') /
        L(Either.import_path)(_).get_or_raise.from_opt(m) |
        m
    )


class EncodeJson(json.JSONEncoder):

    def default(self, obj):
        return (
            obj.json
            if hasattr(obj, 'json') else
            str(obj)
            if isinstance(obj, (uuid.UUID, Path)) else
            super().default(obj)
        )


def json_err(task, data, err):
    return 'error {}coding json: {}\ndata: {}'.format(task, err, data)


def _encode_json(data):
    return EncodeJson().encode(data)


def _decode_json(data):
    return json.loads(data, object_hook=_decode_json_obj)


def _code_json(data, handler, prefix):
    return Try(handler, data).lmap(L(json_err)(prefix, data, _))


def encode_json(data):
    return _code_json(data, _encode_json, 'en')


def decode_json(data):
    return _code_json(data, _decode_json, 'de')

__all__ = ('Record', 'field', 'list_field', 'dfield', 'maybe_field',
           'bool_field', 'any_field', 'encode_json', 'decode_json',
           'int_field', 'uuid_field', 'map_field', 'either_field', 'str_field',
           'float_field')



import abc

from amino import Nothing
from amino.logging import LogError

from ribosome.machine.message_base import message, Message

Nop = message('Nop')
Stop = message('Stop')
Quit = message('Quit')
Done = message('Done')
PlugCommand = message('PlugCommand', 'plug', 'msg')
NvimIOTask = message('NvimIOTask', 'io')
RunTask = message('RunTask', 'task', opt_fields=(('msg', Nothing),))
RunIO = RunTask
RunIOAlg = message('RunIOAlg', 'io', opt_fields=(('msg', Nothing),))
UnitTask = message('UnitTask', 'task', opt_fields=(('msg', Nothing),))
DataTask = message('DataTask', 'cons', opt_fields=(('msg', Nothing),))
Info = message('Info', 'message')
ShowLogInfo = message('ShowLogInfo')
RunIOsParallel = message('RunIOsParallel', 'ios')
RunCorosParallel = message('RunCorosParallel', 'coros')
SubProcessSync = message('SubProcessSync', 'job', 'result')
Warning = message('Warning', 'message')
Debug = message('Debug', 'message')
Coroutine = message('Coroutine', 'coro')
TransitionException = message('TransitionException', 'context', 'exc')


class Error(Message, LogError, fields=('main',), opt_fields=(('prefix', ''),)):

    @property
    def message(self) -> str:
        pre = f'{self.prefix}: ' if self.prefix else ''
        return f'{pre}{self.main}'

    @property
    def full(self) -> str:
        return self.message

    @property
    def short(self) -> str:
        return self.main

__all__ = ('Nop', 'Stop', 'Quit', 'Done', 'Done', 'PlugCommand', 'NvimIOTask', 'RunTask', 'UnitTask', 'DataTask',
           'ShowLogInfo', 'RunIOsParallel', 'RunCorosParallel', 'SubProcessSync', 'RunIO', 'RunIOAlg', 'Error',
           'Warning', 'Debug', 'Coroutine', 'TransitionException', 'Info')



from typing import Callable
from typing import Optional  # NOQA
import abc
import threading
import asyncio
from contextlib import contextmanager
import concurrent.futures

from lenses import Lens

from ribosome.logging import Logging
from ribosome.data import Data
from ribosome.nvim import HasNvim, ScratchBuilder
from ribosome import NvimFacade
from ribosome.machine.message_base import message, Message
from ribosome.machine.base import MachineI, MachineBase
from ribosome.machine.transition import (TransitionResult, Coroutine, may_handle, handle, CoroTransitionResult,
                                         _recover_error, CoroExecutionHandler)
from ribosome.machine.message_base import json_message
from ribosome.machine.helpers import TransitionHelpers
from ribosome.machine.messages import Nop, Done, Quit, PlugCommand, Stop, Error
from ribosome.machine.handler import DynHandlerJob
from ribosome.machine.modular import ModularMachine
from ribosome.machine.transitions import Transitions

import amino
from amino import Maybe, Map, Try, _, L, __, Empty, Just, Either, List, Left
from amino.util.string import red, blue

Callback = message('Callback', 'func')
Envelope = message('Envelope', 'message', 'to')
RunMachine = json_message('RunMachine', 'machine')
KillMachine = message('KillMachine', 'uuid')
RunScratchMachine = json_message('RunScratchMachine', 'machine')
Init = message('Init')
IfUnhandled = message('IfUnhandled', 'msg', 'unhandled')
UpdateRecord = json_message('UpdateRecord', 'tpe', 'name')

short_timeout = 3
medium_timeout = 3
long_timeout = 5
warn_no_handler = True


class AsyncIOBase(Logging, abc.ABC):

    def _init_asyncio(self):
        self._loop = (
            asyncio.new_event_loop()
            if asyncio._get_running_loop() is None  # type: ignore
            else asyncio.get_event_loop()
        )
        self._messages = asyncio.PriorityQueue(loop=self._loop)

    @abc.abstractmethod
    def stop(self, shutdown=True) -> None:
        ...


class AsyncIOThread(threading.Thread, AsyncIOBase):

    def __init__(self) -> None:
        threading.Thread.__init__(self)
        self.done = threading.Event()
        self.quit_now = threading.Event()
        self.running = threading.Event()

    def run(self):
        self._init_asyncio()
        asyncio.set_event_loop(self._loop)
        self.quit_now.clear()
        self.done.clear()
        self.running.set()
        try:
            self._loop.run_until_complete(self._main(self.init))
            self.log.debug('event loop finished in {}'.format(self.title))
        except Exception:
            self.log.exception('while running state machine')
        self.running.clear()
        self.done.set()

    def stop(self, shutdown=True):
        if self.is_alive() and self.running.is_set():
            self.log.debug('stopping machine {}'.format(self.title))
            if shutdown:
                self._stop()
            else:
                self._done()
            self.send(Nop())
            if self.done.wait(long_timeout):
                try:
                    self._loop.close()
                except Exception as e:
                    if amino.development:
                        self.log.caught_exception('stopping {}'.format(self.title))

    def _stop(self):
        self._done()

    def _done(self):
        self.quit_now.set()

    @abc.abstractmethod
    async def _main(self, initial):
        ...

    @property
    def init(self):
        pass


class StateMachineBase(ModularMachine):

    def __init__(self, sub: List[MachineBase]=List(), parent=None, title=None
                 ) -> None:
        self.data = None
        self._messages = None
        self.sub = sub
        ModularMachine.__init__(self, parent, title=None)

    @property
    def init(self) -> Data:
        return self._data_type()

    def start_wait(self):
        self.start()
        self.wait_for_running()

    def _stop(self):
        self.send(Stop())

    def _publish(self, msg):
        return self._messages.put(msg)

    def send(self, msg: Message, prio=0.5):
        self.log.debug('send {} in {}'.format(msg, self.title))
        if self._messages is not None:
            return asyncio.run_coroutine_threadsafe(self._messages.put(msg), self._loop)

    def send_sync(self, msg: Message, wait=None):
        self.send(msg)
        return self.await_state(wait)

    def await_state(self, wait=None):
        (asyncio.run_coroutine_threadsafe(self.join_messages(), self._loop)
         .result(Maybe(wait) | short_timeout))
        return self.data

    def eval_expr(self, expr: str, pre: Callable=lambda a, b: (a, b)):
        sub = self.sub
        sub_map = Map(sub / (lambda a: (a.title, a)))
        data, plugins = pre(self.data, sub_map)
        return Try(eval, expr, None, dict(data=data, plugins=plugins))

    def _send(self, data, msg: Message):
        return (
            Try(self.loop_process, data, msg)
            .value_or(L(TransitionResult.failed)(data, _))
        )

    @may_handle(Nop)
    def _nop(self, data: Data, msg):
        pass

    @may_handle(Stop)
    def _stop_msg(self, data: Data, msg):
        return Quit(), Done().pub.at(1)

    @may_handle(Done)
    def _done_msg(self, data: Data, msg):
        self._done()

    @may_handle(Callback)
    def message_callback(self, data: Data, msg):
        return msg.func(data)

    @may_handle(Coroutine)
    def _couroutine(self, data: Data, msg):
        return msg

    @may_handle(RunMachine)
    def _run_machine(self, data, msg):
        self.sub = self.sub.cat(msg.machine)
        init = msg.options.get('init') | Init()
        return Envelope(init, msg.machine.uuid)

    @may_handle(KillMachine)
    def _kill_machine(self, data, msg):
        self.sub = self.sub.filter_not(_.uuid == msg.uuid)

    @handle(Envelope)
    def message_envelope(self, data, msg):
        return self.sub.find(_.uuid == msg.to) / __.loop_process(data, msg.message)

    @may_handle(IfUnhandled)
    def if_unhandled(self, data, msg):
        result = self._send(data, msg.msg)
        return result if result.handled else self._send(data, msg.unhandled)

    def unhandled(self, data, msg):
        prios = (self.sub // _.prios).distinct.sort(reverse=True)
        step = lambda z, a: z if z.handled else self._fold_sub(z.data, msg, a)
        return Just(prios.fold_left(TransitionResult.unhandled(data))(step))

    def _fold_sub(self, data, msg, prio=None):
        ''' send **msg** to all sub-machines, passing the transformed
        data from each machine to the next and accumulating published
        messages.
        '''
        send = lambda z, s: z.accum(s.loop_process(z.data, msg, prio))
        return self.sub.fold_left(TransitionResult.unhandled(data))(send)

    @contextmanager
    def transient(self):
        self.start()
        self.wait_for_running()
        self.send(Nop())
        yield self
        self.stop()

    async def _step(self):
        self.data = await self._process_one_message(self.data)

    async def _process_one_message(self, data):
        msg = await self._messages.get()
        sent = self._send(data, msg)
        result = (
            self._failed_transition(msg, sent)
            if sent.failure else
            await self._successful_transition(data, msg, sent)
            if sent.handled else
            self._unhandled_message(msg, sent)
        )
        self._messages.task_done()
        return result.data

    async def _successful_transition(self, data, msg, sent):
        result = (
            await self._execute_coro_result(data, msg, sent)
            if isinstance(sent, CoroTransitionResult) else
            sent
        )
        for pub in result.pub:
            await self._publish(pub)
        return result

    async def _execute_coro_result(self, data: Data, msg: Message, ctr: CoroTransitionResult) -> None:
        try:
            job = DynHandlerJob(self, data, msg, CoroExecutionHandler(self, 'coroutine result', msg, None, 1.0, True),
                                self._data_type)
        except Exception as e:
            return TransitionResult.failed(data, f'failed to create coro handler job: {e}')
        try:
            result0 = await ctr.coro.coro
            result = job.dispatch_transition_result(_recover_error(self, result0))
        except Exception as e:
            return job.handle_transition_error(e)
        else:
            if result.resend:
                msg = 'Cannot resend {} from coro {}, use .pub on messages'
                self.log.warn(msg.format(result.resend, ctr.coro))
            return result

    def _failed_transition(self, msg, sent: TransitionResult):
        msg_name = type(msg).__name__
        def log_error() -> None:
            errmsg = sent.error_message
            self.log.error(Error(f'''{red('error')} handling {blue(msg_name)}''', prefix=errmsg))
        def log_exc(e: Exception) -> None:
            self.log.caught_exception_error(f'handling {blue(msg_name)}', e)
        def log_verbose() -> None:
            sent.exception / log_exc
        try:
            log_error()
            if amino.development:
                log_verbose()
        except Exception as e:
            self.log.error(f'error in `failed_transition`: {e}')
        return sent

    def _unhandled_message(self, msg, sent):
        log = self.log.warning if warn_no_handler else self.log.debug
        log('{}: no handler for {}'.format(self.title, msg))
        return sent

    async def join_messages(self):
        await self._messages.join()


class StateMachine(StateMachineBase, AsyncIOThread):

    def __init__(self, *a, **kw) -> None:
        AsyncIOThread.__init__(self)
        StateMachineBase.__init__(self, *a, **kw)

    def wait_for_running(self):
        self.running.wait(medium_timeout)

    async def _main(self, data):
        self.data = data
        while not self.quit_now.is_set():
            await self._step()


class UnloopedStateMachine(StateMachineBase, AsyncIOBase):

    def start(self):
        self.data = self.init
        self._init_asyncio()

    def stop(self, shutdown=True) -> None:
        pass

    def wait_for_running(self):
        pass

    def process_messages(self):
        return self._loop.run_until_complete(self._process_messages())

    async def _process_messages(self):
        while not self._messages.empty():
            await self._step()

    def send_sync(self, msg, timeout=medium_timeout):
        self.log.debug('send {} in {}'.format(msg, self.title))
        done = concurrent.futures.Future()  # type: concurrent.futures.Future
        async def go():
            if self._messages is not None:
                await self._messages.put(msg)
            result = await self._process_messages()
            done.set_result(result)
        if self._loop.is_running():
            self._loop.call_soon_threadsafe(go())
        else:
            self._loop.run_until_complete(go())
        return done.result(timeout)

    def send(self, msg):
        threading.Thread(target=self.send_sync, args=(msg,)).start()


class PluginStateMachine(MachineI):

    def __init__(self, plugins: List[str]) -> None:
        self.sub = plugins.flat_map(self.start_plugin)

    def start_plugin(self, name: str):
        def report(errs):
            msg = 'invalid {} plugin module "{}": {}'
            self.log.error(msg.format(self.title, name, errs))
        return (self._find_plugin(name) // self._inst_plugin).leffect(report)

    def _find_plugin(self, name: str):
        mods = List(
            Either.import_module(name),
            Either.import_module('{}.plugins.{}'.format(self.title, name))
        )
        errors = mods.filter(_.is_left) / _.value
        return mods.find(_.is_right) | Left(errors)

    def _inst_plugin(self, mod):
        return (
            Try(getattr(mod, 'Plugin'), self.vim, self)
            if hasattr(mod, 'Plugin')
            else Left('module does not define class `Plugin`')
        )

    def plugin(self, title):
        return self.sub.find(_.title == title)

    def plug_command(self, plug_name: str, cmd_name: str, args: list=[],
                     sync=False):
        sender = self.send_sync if sync else self.send
        plug = self.plugin(plug_name)
        cmd = plug.flat_map(lambda a: a.command(cmd_name, List(args)))
        plug.ap2(cmd, PlugCommand) % sender

    def plug_command_sync(self, *a, **kw):
        return self.plug_command(*a, sync=True, **kw)

    @may_handle(PlugCommand)
    def _plug_command(self, data, msg):
        self.log.debug(
            'sending command {} to plugin {}'.format(msg.msg, msg.plug.title))
        return msg.plug.process(data, msg.msg)


# FIXME why is the title param ignored?
class RootMachineBase(PluginStateMachine, HasNvim, Logging):

    def __init__(self, vim: NvimFacade, plugins: List[str]=List(), title=None) -> None:
        HasNvim.__init__(self, vim)
        PluginStateMachine.__init__(self, plugins)

    @property
    def title(self):
        return 'ribosome'

    @handle(RunScratchMachine)
    def _scratch_machine(self, data, msg):
        return (
            ScratchBuilder(**(msg.options - 'init'))
            .build
            .unsafe_perform_io(self.vim) /
            L(msg.machine)(_, self) /
            L(RunMachine)(_, msg.options)
        )


class RootMachine(StateMachine, RootMachineBase):

    def __init__(self, vim: NvimFacade, plugins: List[str]=List(), title=None
                 ) -> None:
        StateMachine.__init__(self, title=title)
        RootMachineBase.__init__(self, vim, plugins)


class UnloopedRootMachine(UnloopedStateMachine, RootMachineBase):

    def __init__(self, vim: NvimFacade, plugins: List[str]=List(), title=None
                 ) -> None:
        UnloopedStateMachine.__init__(self, title=title)
        RootMachineBase.__init__(self, vim, plugins)


class SubMachine(ModularMachine, TransitionHelpers):

    def new_state(self):
        pass


class SubTransitions(Transitions, TransitionHelpers):

    def _state(self, data):
        return data.sub_state(self.name, self.new_state)

    @property
    def state(self):
        return self._state(self.data)

    def _with_sub(self, data, state):
        return data.with_sub_state(self.name, state)

    def with_sub(self, state):
        return self._with_sub(self.data, state)

    @property
    def new_state(self):
        return self.machine.new_state

    @property
    def options(self):
        return getattr(self.msg, 'options', Map())

    @handle(UpdateRecord)
    def message_update_record(self):
        return (
            self.record_lens(self.msg.tpe, self.msg.name) /
            __.modify(__.update_from_opt(self.msg.options)) /
            self.with_sub
        )

    def record_lens(self, tpe, name) -> Maybe[Lens]:
        return Empty()

__all__ = ('Message', 'StateMachine', 'PluginStateMachine')



import abc
from typing import Union, Any, Callable

import neovim

from amino import env

from ribosome.nvim import NvimFacade
from ribosome.machine import StateMachine
from ribosome.logging import nvim_logging, Logging
from ribosome.request import msg_command, msg_function, command
from ribosome.machine.base import ShowLogInfo
from ribosome.machine.scratch import Mapping
from ribosome.rpc import setup_rpc


class NvimPlugin(Logging):
    name: str = None
    prefix: str = None

    def __init_subclass__(cls: type, name: str=None, prefix: str=None) -> None:
        if name is not None:
            setup_plugin(cls, name, prefix or name)

    def __init__(self, nvim: Union[NvimFacade, neovim.Nvim]) -> None:
        self.vim = NvimFacade(nvim, self.name) if isinstance(nvim, neovim.Nvim) else nvim
        self.setup_logging()
        if 'SETUP_RPC' in env:
            setup_rpc(self.vim, self.name, type(self))

    def setup_logging(self) -> None:
        self.file_log_handler = nvim_logging(self.vim)

    @property
    def loop(self) -> Any:
        return self.vim.loop

    @property
    def name(self) -> str:
        return type(self) or 'ribosome'

    @abc.abstractmethod
    def start_plugin(self) -> None:
        ...

    def set_log_level(self, level: str) -> None:
        self.file_log_handler.setLevel(level)


class NvimStatePlugin(NvimPlugin):

    @abc.abstractproperty
    def state(self) -> StateMachine:
        ...

    @property
    def default_sync(self) -> bool:
        return False


def setup_plugin(cls: type, name: str, prefix: str) -> None:
    def name_handler(suf: str, handler: Callable[[str], Callable[..., None]]) -> None:
        n = f'{name}_{suf}'
        setattr(cls, n, handler(n))
    def handler(suf: str, handler: Callable[[str], Callable[..., None]]) -> None:
        n = f'{prefix}_{suf}'
        setattr(cls, n, handler(n))
    def msg_cmd(suf: str, msg: type) -> None:
        handler(suf, lambda n: msg_command(msg, name=n)(lambda: None))
    def msg_fun(suf: str, msg: type) -> None:
        handler(suf, lambda n: msg_function(msg, name=n)(lambda: None))
    cls.name = name
    cls.prefix = prefix
    msg_cmd('show_log_info', ShowLogInfo)
    handler('log_level', lambda n: command(name=n)(lambda self, *a, **kw: self.set_log_level(*a, **kw)))
    msg_fun('mapping', Mapping)
    name_handler('start', lambda n: command(sync=True, name=n)(lambda self, *a, **kw: self.start_plugin(*a, **kw)))

__all__ = ('NvimPlugin', 'NvimStatePlugin')



rplugin_template = '''
import neovim
import os
import sys
import logging
from pathlib import Path

import amino

amino.development = True

import amino.logging
from ribosome.logging import ribosome_root_logger

logfile = Path(os.environ['RIBOSOME_LOG_FILE'])
fmt = os.environ.get('RIBOSOME_FILE_LOG_FMT')
amino.logging.amino_root_file_logging(level=amino.logging.TEST, logfile=logfile, fmt=fmt)

pkg_dir = os.environ.get('RIBOSOME_PKG_DIR')
if pkg_dir:
    sys.path.insert(0, pkg_dir)

try:
    from {plugin_module} import {plugin_class}
except Exception as e:
    ribosome_root_logger.caught_exception('importing {plugin_module}.{plugin_class}', e)
    raise

@neovim.plugin
class Plugin({plugin_class}):
    pass
'''

__all__ = ('rplugin_template',)



import os
import logging
from typing import TypeVar, Callable

from toolz import merge

import amino
from amino.lazy import lazy
import amino.logging
from amino.logging import (amino_logger, init_loglevel, amino_root_file_logging, DDEBUG, print_log_info, VERBOSE,
                           LazyRecord)
from amino import Path, Logger

import ribosome  # noqa
from ribosome import options

A = TypeVar('A')


class NvimHandler(logging.Handler):

    def __init__(self, vim: 'ribosome.NvimFacade') -> None:
        self.vim = vim
        self.dispatchers = {
            VERBOSE: self.vim.echom,
            logging.INFO: self.vim.echo,
            logging.WARN: self.vim.echowarn,
            logging.ERROR: self.vim.echoerr,
            logging.CRITICAL: self.vim.echoerr,
        }
        super().__init__()

    def emit(self, record: LazyRecord) -> None:
        dispatcher = self.dispatchers.get(record.levelno, self.vim.echom)
        dispatcher(record.short())


class NvimFilter(logging.Filter):

    def filter(self, record: logging.LogRecord) -> bool:
        return record.exc_info is None


nvim_filter = NvimFilter()
ribo_log = log = ribosome_root_logger = amino_logger('nvim')
_nvim_logging_initialized = False


def ribosome_logger(name: str) -> Logger:
    return ribosome_root_logger.getChild(name)


def nvim_logging(vim: 'ribosome.NvimFacade', level: int=logging.INFO, file_kw: dict=dict()) -> logging.Handler:
    global _nvim_logging_initialized
    if not _nvim_logging_initialized:
        if level is None and not amino.development:
            level = logging.INFO
        handler = NvimHandler(vim)
        handler.addFilter(nvim_filter)
        ribosome_root_logger.addHandler(handler)
        init_loglevel(handler, VERBOSE)
        def file_log(prefix: str) -> logging.Handler:
            level = (
                DDEBUG
                if options.development and options.spec else
                options.file_log_level.value | logging.INFO
            )
            logfile = Path('{}_ribo_{}'.format(prefix, os.getpid()))
            fmt = options.file_log_fmt.value / (lambda fmt: dict(fmt=fmt)) | dict()
            kw = merge(
                file_kw,
                dict(level=level, logfile=logfile),
                fmt
            )
            return amino_root_file_logging(**kw)
        _nvim_logging_initialized = True
        logfile = options.nvim_log_file.value | (Path.home() / '.cache' / 'ribosome')
        return file_log(logfile)


class Logging(amino.logging.Logging):

    @lazy
    def _log(self) -> Logger:
        return ribosome_logger(self.__class__.__name__)


def pr(a: A) -> A:
    v = log.verbose
    v(a)
    return a


def pv(a: A) -> A:
    v = log.verbose
    v(str(a))
    return a


def print_ribo_log_info(out: Callable[[str], None]) -> None:
    print_log_info(out)
    out(str(options.development))
    out(str(options.spec))
    out(str(options.file_log_level))
    out(str(options.file_log_fmt))
    out(str(options.nvim_log_file))

__all__ = ('ribosome_logger', 'nvim_logging', 'Logging', 'pr', 'ribo_log')



import abc
import time
from typing import Any, Sequence, TypeVar, Generic, Type, Tuple
import asyncio

import amino
from amino import Maybe, _, List, Map, Just, Either, L, __, Right
from amino.task import Task
from amino.state import StateT, EvalState, MaybeState, EitherState, IdState
from amino.tc.optional import Optional
from amino.id import Id
from amino.util.string import blue
from amino.dispatch import dispatch_alg
from amino.tc.base import TypeClass
from amino.func import flip

from ribosome.logging import Logging
from ribosome.machine.message_base import Message, Publish
from ribosome.machine.transition import (Handler, TransitionResult, CoroTransitionResult, StrictTransitionResult,
                                         TransitionFailed, Coroutine, MachineError, Error)
from ribosome.machine.messages import RunTask, UnitTask, DataTask, Nop
from ribosome.machine.interface import MachineI
from ribosome.data import Data
from ribosome.machine.trans import TransAction, Transit, Propagate, Unit, TransFailure


def is_seq(a):
    return isinstance(a, Sequence)


def is_message(a):
    return isinstance(a, Message)


class Handlers(Logging):

    def __init__(self, prio: int, handlers: Map[type, Handler]) -> None:
        self.prio = prio
        self.handlers = handlers

    def handler(self, msg):
        return self.handlers.get(type(msg))


M = TypeVar('M', bound=MachineI)
D = TypeVar('D', bound=Data)
Msg = TypeVar('Msg', bound=Message)
R = TypeVar('R')


def create_result(data: D, msgs: List[Message]) -> TransitionResult:
    pub, resend = msgs.split_type(Publish)
    pub_msgs = pub.map(_.message)
    return StrictTransitionResult(data=data, pub=pub_msgs, resend=resend)


def create_failure(data: D, desc: str, error: str) -> TransitionResult:
    msg = f'{desc}: {error}'
    return TransitionResult.failed(data, msg)


class HandlerJob(Generic[M, D], Logging):

    def __init__(self, machine: M, data: D, msg: Message, handler: Handler, data_type: Type[D]) -> None:
        self.machine = machine
        self.data = data
        self.msg = msg
        self.handler = handler
        self.data_type = data_type
        self.start_time = time.time()

    @abc.abstractmethod
    def run(self) -> TransitionResult:
        ...

    @property
    def trans_desc(self) -> str:
        return blue(f'{self.machine.title}.{self.handler.name}')


class DynHandlerJob(HandlerJob):

    def run(self) -> TransitionResult:
        result = self._execute_transition(self.handler, self.data, self.msg)
        return self.dispatch_transition_result(result)

    def _execute_transition(self, handler, data, msg):
        try:
            return handler.run(data, msg)
        except TransitionFailed as e:
            raise
        except Exception as e:
            return self.handle_transition_error(e)

    def handle_transition_error(self, e):
        if amino.development:
            err = 'transition "{}" failed for {} in {}'
            self.log.caught_exception(err.format(self.handler.name, self.msg, self.machine.title), e)
            raise TransitionFailed(str(e)) from e
        return self.failure_result(str(e))

    def failure_result(self, err: str) -> Maybe[StrictTransitionResult]:
        return Just(TransitionResult.failed(self.data, err))

    def dispatch_transition_result(self, result):
        return (
            result /
            self.process_result |
            TransitionResult.empty(self.data)
        )

    def process_result(self, res0: Any) -> TransitionResult:
        if isinstance(res0, Coroutine):
            return CoroTransitionResult(data=self.data, coro=res0)
        elif isinstance(res0, TransitionResult):
            return res0
        elif isinstance(res0, self.data_type):
            return TransitionResult.empty(res0)
        elif isinstance(res0, StateT):
            result = self.transform_state(res0)
        elif is_message(res0) or not is_seq(res0):
            result = List(res0)
        else:
            result = res0
        datas, rest = List.wrap(result).split_type(self.data_type)
        trans = rest / self.transform_result
        msgs, rest = trans.split_type(Message)
        if rest:
            tpl = 'invalid transition result parts for {} in {}: {}'
            msg = tpl.format(self.msg, self.machine.title, rest)
            if amino.development:
                raise MachineError(msg)
            else:
                self.log.error(msg)
        new_data = datas.head | self.data
        return create_result(new_data, msgs)

    def transform_result(self, result):
        if asyncio.iscoroutine(result) or isinstance(result, asyncio.futures.Future):
            return Coroutine(result).pub
        elif isinstance(result, Task):
            return RunTask(result, Just(self.msg))
        elif isinstance(result, RunTask):
            return RunTask(result.task, Just(self.msg))
        elif isinstance(result, UnitTask):
            return UnitTask(result.task, Just(self.msg))
        elif isinstance(result, DataTask):
            return DataTask(result.cons, Just(self.msg))
        else:
            return result

    def transform_state(self, res0: StateT):
        res1 = res0.run(self.data)
        if isinstance(res0, EvalState):
            (data, result) = res1._value()
        elif isinstance(res0, MaybeState):
            (data, result) = res1.get_or_else((self.data, Nop()))
        elif isinstance(res0, EitherState):
            (data, result) = res1.value_or(lambda a: (self.data, Error(str(a))))
        elif isinstance(res0, IdState):
            (data, result) = res1.value
        else:
            return List(Error(f'invalid effect for transition result `State`: {res0.tpe}#{res1}'))
        r2 = (
            Nop()
            if result is None else
            result.get_or_else(Nop())
            if Optional.exists(type(result)) else
            result
        )
        return (
            r2.cons(data)
            if isinstance(r2, List) else
            List(data, r2)
        )


A = TypeVar('A')


class UnpackTransState(Generic[A], TypeClass[A]):

    @abc.abstractmethod
    def unpack(self, a: A, data: D) -> Either[Message, Tuple[D, List[Message]]]:
        ...


class UnpackIdState(Generic[D], UnpackTransState[IdState[D, List[Message]]], tpe=IdState):

    def unpack(self, a: IdState[D, List[Message]], data: D) -> Either[Message, Tuple[D, List[Message]]]:
        return Right(a.run(data).value)


class UnpackEitherState(Generic[D], UnpackTransState[EitherState[D, List[Message]]], tpe=EitherState):

    def unpack(self, a: EitherState[D, List[Message]], data: D) -> Either[Message, Tuple[D, List[Message]]]:
        return a.run(data)


class AlgResultValidator(Logging):

    def __init__(self, desc: str) -> None:
        self.desc = desc

    @property
    def validate(self) -> TransitionResult:
        return dispatch_alg(self, TransAction, 'validate_')

    def failure(self, data: D, error: str) -> TransitionResult:
        return create_failure(data, self.desc, error)

    def validate_transit(self, action: Transit, data: D) -> TransitionResult:
        return (
            UnpackTransState.e_for(action.trans)
            .flat_map(__.unpack(action.trans, data))
            .map2(flip)
            .map2(self.validate)
            .value_or(L(self.failure)(data, _))
        )

    def validate_propagate(self, action: Propagate, data: D) -> TransitionResult:
        return create_result(data, action.messages)

    def validate_unit(self, action: Unit, data: D) -> TransitionResult:
        return create_result(data, action.messages)

    def validate_trans_failure(self, action: TransFailure, data: D) -> TransitionResult:
        return self.failure(data, action.message)


class AlgHandlerJob(HandlerJob):

    def run(self) -> TransitionResult:
        try:
            r = self.handler.execute(self.data, self.msg)
        except Exception as e:
            return self.exception(e)
        else:
            return AlgResultValidator(self.trans_desc).validate(r, self.data)

    def exception(self, e: Exception) -> StateT[Id, D, R]:
        if amino.development:
            err = f'transitioning {self.trans_desc}'
            self.log.caught_exception(err, e)
            raise TransitionFailed(str(e)) from e
        return create_failure(self.data, self.trans_desc, f'exception raised: {e}')

__all__ = ('Handlers', 'HandlerJob', 'AlgHandlerJob', 'DynHandlerJob')



from kallikrein import Expectation
from kallikrein.matchers.maybe import be_just
from kallikrein.matchers.end_with import end_with

from amino import Either, Right

from integration._support.handler import HandlerSpecPlugin
from integration._support.base import IntegrationSpecBase


class HandlerSpec(IntegrationSpecBase):
    '''typed transition handler
    nested handlers $nested
    unit handler $unit
    EitherState handler $est
    '''

    @property
    def plugin_class(self) -> Either[str, type]:
        return Right(HandlerSpecPlugin)

    def nested(self) -> Expectation:
        self.vim.cmd('HandlerStart')
        self.vim.cmd_sync('Msg')
        return self._log_line(-1, be_just('nothing'))

    def unit(self) -> Expectation:
        self.vim.cmd('HandlerStart')
        self.vim.cmd_sync('Unit')
        return self._log_line(-1, be_just('unit'))

    def est(self) -> Expectation:
        self.vim.cmd('HandlerStart')
        self.vim.cmd_sync('Est')
        return self._log_line(-1, be_just(end_with(': est')))

__all__ = ('HandlerSpec',)



from ribosome.test.integration.klk import PluginIntegrationKlkSpec


class IntegrationSpecBase(PluginIntegrationKlkSpec):

    def setup(self):
        self.log_format = '{message}'
        super().setup()

__all__ = ('IntegrationSpecBase',)



import inspect
from typing import TypeVar, Generic, Type

from amino import List, L, _

from ribosome.machine.transitions import Transitions
from ribosome.machine.base import MachineBase
from ribosome.machine.transition import WrappedHandler
from ribosome.machine.message_base import _machine_attr

T = TypeVar('T', bound=Transitions)


class ModularMachine(Generic[T], MachineBase):
    Transitions: Type[T] = Transitions

    @property
    def _handlers(self):
        methods = inspect.getmembers(self.Transitions, lambda a: hasattr(a, _machine_attr))
        handlers = (
            List.wrap(methods)
            .map2(L(WrappedHandler.create)(self, _, _, self.Transitions))
        )
        return handlers + super()._handlers

__all__ = ('ModularMachine',)



import abc
from typing import Callable, TypeVar, Type, Generic, Coroutine
import functools

from amino import Either, List, Lists, IO, Id, L, _, Maybe
from amino.state import IdState, StateT, EitherState
from amino.util.string import red, green, ToStr
from amino.dispatch import dispatch_alg
from amino.algebra import AlgebraMeta, Algebra

from ribosome.machine.message_base import Message, default_prio, _machine_attr, _message_attr, _prio_attr, _dyn_attr
from ribosome.data import Data
from ribosome.machine.interface import MachineI
from ribosome.machine.messages import RunIOAlg, Error
from ribosome.logging import Logging

M = TypeVar('M', bound=MachineI)
D = TypeVar('D', bound=Data)
Msg = TypeVar('Msg', bound=Message)
R = TypeVar('R')
N = TypeVar('N')
O = TypeVar('O')
G = TypeVar('G')


class TransAction(Algebra, base=True):

    def __init__(self, messages: List[Message]) -> None:
        self.messages = messages


class Transit(Generic[D], TransAction):

    def __init__(self, trans: StateT[Id, D, 'Propagate']) -> None:
        super().__init__(List())
        self.trans = trans

    def _arg_desc(self) -> List[str]:
        return List(str(self.trans))


class Propagate(TransAction):

    @staticmethod
    def one(msg: Message) -> TransAction:
        return Propagate(List(msg))

    def _arg_desc(self) -> List[str]:
        return self.messages / str


class Unit(TransAction):

    def __init__(self) -> None:
        super().__init__(List())

    def _arg_desc(self) -> List[str]:
        return List()


class TransFailure(TransAction):

    def __init__(self, message: str) -> None:
        super().__init__(List())
        self.message = message

    def _arg_desc(self) -> List[str]:
        return List(self.message)


class TransStep(Generic[R], ToStr, metaclass=AlgebraMeta, base=True):

    def __init__(self, data: R) -> None:
        self.data = data

    @property
    def strict(self) -> bool:
        return isinstance(self, Strict)

    @property
    def error(self) -> bool:
        return isinstance(self, TransEffectError)

    def _arg_desc(self) -> List[str]:
        return List(str(self.data))


class Strict(Generic[R], TransStep[R]):
    pass


class Lift(Generic[R], TransStep[R]):
    pass


class TransEffectError(TransStep[str]):
    pass


class TransEffect(Generic[R], abc.ABC, Logging):

    @abc.abstractproperty
    def tpe(self) -> Type[R]:
        ...

    @abc.abstractmethod
    def extract(self, data: R, tail: List['TransEffect'], in_state: bool) -> TransStep:
        ...

    def run(self, data: R, tail: List['TransEffect'], in_state: bool) -> TransStep:
        return (
            self.extract(data, tail, in_state)
            if isinstance(data, self.tpe) else
            TransEffectError(f'result {red(data)} does not have type {green(self.tpe.__qualname__)}')
        )


def cont(tail: List[TransEffect], in_state: bool, f: Callable[[Callable[[R], TransStep]], TransStep]
         ) -> Maybe[TransStep]:
    return tail.detach_head.map2(lambda h, t: f(lambda a: h.run(a, t, in_state)))


class TransEffectEither(Generic[R], TransEffect[Either[str, R]]):

    @property
    def tpe(self) -> Type[Either[str, R]]:
        return Either

    def extract(self, data: Either[R, N], tail: List[TransEffect], in_state: bool) -> TransStep:
        nested = data.map(lambda a: cont(tail, in_state, lambda run: Lift(run(a))) | a)
        return Lift(nested.value_or(L(Error)(_) >> Propagate.one))


class TransEffectStateT(Generic[G, D, R], TransEffect[StateT[G, D, R]]):

    def check(self, result: TransEffect[R]) -> None:
        return (
            TransEffectError(f'`State` trans result `{result}` is not a `TransStep`')
            if not isinstance(result, TransStep) else
            TransEffectError('found nested `State` in transition result')
            if result.strict else
            TransEffectError(result.data)
            if result.error else
            result.data
        )


class TransEffectIdState(Generic[D, R], TransEffectStateT[Id, D, R]):

    @property
    def tpe(self) -> Type[IdState[D, R]]:
        return IdState

    def extract(self, data: IdState, tail: List[TransEffect], in_state: bool) -> IdState:
        return (
            TransEffectError('cannot nest `State` in transition result')
            if in_state else
            Strict(cont(tail, True, lambda run: data.map(lambda inner: self.check(run(inner)))) | data)
        )


class TransEffectEitherState(Generic[D, R], TransEffectStateT[Either, D, R]):

    @property
    def tpe(self) -> Type[EitherState[D, R]]:
        return EitherState

    def extract(self, data: EitherState[D, R], tail: List[TransEffect], in_state: bool) -> IdState:
        return (
            TransEffectError('cannot nest `State` in transition result')
            if in_state else
            Strict(cont(tail, True, lambda run: data.map(lambda inner: self.check(run(inner)))) | data)
        )


class TransEffectIO(Generic[R], TransEffect[IO[R]]):

    @property
    def tpe(self) -> Type[IO[R]]:
        return IO

    def extract(self, data: IO[R], tail: List[TransEffect], in_state: bool) -> Either[R, N]:
        io = cont(tail, False, data.map) | data
        return Lift(Propagate.one(RunIOAlg(io.map(L(lift)(_, in_state)))))


class TransEffectCoro(TransEffect):

    @property
    def tpe(self) -> Type[Coroutine]:
        return Coroutine

    def extract(self, data: IO[R], tail: List[TransEffect], in_state: bool) -> Either[R, N]:
        ...


class TransEffectSingleMessage(TransEffect[Message]):

    @property
    def tpe(self) -> Type[Message]:
        return Message

    def extract(self, data: Message, tail: List[TransEffect], in_state: bool) -> Either[R, N]:
        return Lift(Propagate.one(data)) if tail.empty else TransEffectError('cannot apply trans effects to Message')


class TransEffectMessages(TransEffect[List[Message]]):

    @property
    def tpe(self) -> Type[List[Message]]:
        return List

    def extract(self, data: List[Message], tail: List[TransEffect], in_state: bool) -> Either[R, N]:
        return Lift(Propagate(data)) if tail.empty else TransEffectError('cannot apply trans effects to Messages')


class TransEffectUnit(TransEffect[None]):

    @property
    def tpe(self) -> Type[Message]:
        return type(None)

    def extract(self, data: Message, tail: List[TransEffect], in_state: bool) -> Either[R, N]:
        return Lift(Unit()) if tail.empty else TransEffectError('cannot apply trans effects to unit')


e = TransEffectEither()
st = TransEffectIdState()
est = TransEffectEitherState()
io = TransEffectIO()
coro = TransEffectCoro()
single = TransEffectSingleMessage()
strict = TransEffectMessages()
none = TransEffectUnit()


class Lifter(Logging):

    def lift_lift(self, res: Lift, in_state: bool) -> TransAction:
        return lift(res.data, in_state)

    def lift_strict(self, res: Strict, in_state: bool) -> TransAction:
        return Transit(res.data / L(lift)(_, True))

    def lift_trans_effect_error(self, res: TransEffectError, in_state: bool) -> TransAction:
        return TransFailure(res.data)
        # return prop if in_state else Transit(IdState.pure(prop))

    def lift_res(self, res: R, in_state: bool) -> TransAction:
        return (
            res
            if isinstance(res, TransAction) else
            TransFailure(f'transition did not produce `TransAction`: {{red(res)}}')
        )


_lifter = Lifter()
lift = dispatch_alg(_lifter, TransStep, 'lift_', _lifter.lift_res)


def extract(output: O, effects: List[TransEffect]) -> TransAction:
    trans_result = cont(effects, False, lambda f: f(output)) | output
    return lift(trans_result, False)


def decorate(transition, msg_type, prio) -> None:
    setattr(transition, _machine_attr, True)
    setattr(transition, _message_attr, msg_type)
    setattr(transition, _prio_attr, prio)
    setattr(transition, _dyn_attr, False)
    return transition


def base(msg_type: Type[Msg], *effects: TransEffect, prio: float=default_prio
         ) -> Callable[[Callable[[M], R]], Callable[[M], TransAction]]:
    def add_handler(func: Callable[[M], R]):
        @functools.wraps(func)
        def transition(m, *a) -> None:
            return extract(func(m, *a), Lists.wrap(effects))
        return decorate(transition, msg_type, prio)
    return add_handler


def unit(msg_type: Type[Msg], *effects: TransEffect, prio: float=default_prio
         ) -> Callable[[Callable[[M], R]], Callable[[M], TransAction]]:
    return base(msg_type, *effects, none, prio=prio)


def one(msg_type: Type[Msg], *effects: TransEffect, prio: float=default_prio
        ) -> Callable[[Callable[[M], R]], Callable[[M], TransAction]]:
    return base(msg_type, *effects, single, prio=prio)


def multi(msg_type: Type[Msg], *effects: TransEffect, prio: float=default_prio
          ) -> Callable[[Callable[[M], R]], Callable[[M], TransAction]]:
    return base(msg_type, *effects, strict, prio=prio)


def relay(msg_type: Type[Msg], prio: float=default_prio) -> Callable[[Callable[[M], R]], Callable[[M], TransAction]]:
    def add_handler(func: Callable[[M], R]):
        return decorate(func, msg_type, prio)
    return add_handler

__all__ = ('multi', 'e', 'st', 'io', 'coro', 'single', 'one', 'relay')



import os
import abc
import json
import asyncio
from functools import wraps
from threading import Thread
from contextlib import contextmanager
from typing import Any

import neovim
from neovim.api import Nvim

from amino.test import fixture_path, temp_dir, temp_file

from amino import List, Maybe, Either, Left, __, Map, env, Path
from amino.lazy import lazy
from amino.test.path import base_dir, pkg_dir
from amino.test.spec import IntegrationSpecBase as AminoIntegrationSpecBase

import ribosome
from ribosome.logging import Logging
from ribosome import NvimFacade
from ribosome.nvim import AsyncVimProxy
from ribosome.test.fixtures import rplugin_template


class IntegrationSpecBase(AminoIntegrationSpecBase):

    def setup(self) -> None:
        AminoIntegrationSpecBase.setup(self)
        AsyncVimProxy.allow_async_relay = False

    def teardown(self) -> None:
        AminoIntegrationSpecBase.teardown(self)


class VimIntegrationSpecI(abc.ABC):

    @abc.abstractproperty
    def _log_out(self) -> List[str]:
        ...


class VimIntegrationSpec(VimIntegrationSpecI, IntegrationSpecBase, Logging):

    def __init__(self) -> None:
        IntegrationSpecBase.__init__(self)
        self.tmux_nvim = 'RIBOSOME_TMUX_SPEC' in env
        self._tmux_pane = None
        self.keep_tmux_pane = False
        self.vimlog = temp_dir('log') / 'vim'
        self._cmdline = ('nvim', '-V{}'.format(self.vimlog), '-n', '-u', 'NONE')
        self.log_format = '{levelname} {name}:{message}'

    def setup(self) -> None:
        IntegrationSpecBase.setup(self)
        env['RIBOSOME_SPEC'] = 1
        env['AMINO_DEVELOPMENT'] = 1
        env['RIBOSOME_PKG_DIR'] = str(pkg_dir())
        self._debug = 'RIBOSOME_DEVELOPMENT' in env
        self.logfile = temp_dir('log') / self.__class__.__name__
        self.logfile.touch()
        os.environ['RIBOSOME_LOG_FILE'] = str(self.logfile)
        os.environ['RIBOSOME_FILE_LOG_FMT'] = self.log_format
        self._pre_start_neovim()
        self._start_neovim()
        self._post_start_neovim()
        self._pre_start()
        self._start_plugin()
        self._post_start()

    def _start_neovim(self) -> None:
        asyncio.get_child_watcher().attach_loop(asyncio.get_event_loop())
        if self.tmux_nvim:
            self._start_neovim_tmux()
        else:
            self._start_neovim_embedded()
        self.vim = self._nvim_facade(self.neovim)

    def _start_neovim_embedded(self) -> None:
        ''' start an embedded vim session that loads no init.vim.
        **self.vimlog** is set as log file. aside from being convenient,
        this is crucially necessary, as the first use of the session
        will block if stdout is used for output.
        '''
        argv = self._cmdline + ('--embed',)
        self.neovim = neovim.attach('child', argv=argv)

    @lazy
    def nvim_socket(self) -> str:
        return str(temp_dir('nvim_sockets') / List.random_string())

    @property
    def project_path(self) -> str:
        return str(base_dir().parent)

    @lazy
    def _tmux_window(self) -> Any:
        try:
            import libtmux
        except ImportError:
            raise Exception('install libtmux to run nvim in a tmux pane')
        else:
            server = libtmux.Server()
            session = next(s for s in server.sessions
                           if int(s['session_attached']) >= 1)
            return session.attached_window

    def _start_neovim_tmux(self) -> None:
        from libtmux import Pane
        global_path = env['PYTHONPATH'] | ''
        path = '{}:{}'.format(self.project_path, global_path)
        tmux_env = (
            'env',
            'NVIM_LISTEN_ADDRESS={}'.format(self.nvim_socket),
            'PYTHONPATH={}'.format(path),
            'RIBOSOME_LOG_FILE={}'.format(self.logfile),
            'RIBOSOME_FILE_LOG_FMT=\'{}\''.format(self.log_format),
        )
        cmd = tuple(tmux_env + self._cmdline)
        out = self._tmux_window.cmd(
            'split-window', '-d', '-P', '-F#{pane_id}', *cmd).stdout
        self._tmux_pane = Pane(self._tmux_window, pane_id=out[0])
        self.neovim = neovim.attach('socket', path=self.nvim_socket)
        self.neovim.command('python3 sys.path.insert(0, \'{}\')'.format(path))

    def _cleanup_tmux(self) -> None:
        if self._tmux_pane is not None and not self.keep_tmux_pane:
            self._tmux_pane.cmd('kill-pane')

    def _nvim_facade(self, vim: Nvim) -> NvimFacade:
        return NvimFacade(vim, self._prefix)

    @property
    def _prefix(self) -> str:
        return ''

    def teardown(self) -> None:
        IntegrationSpecBase.teardown(self)
        if self._debug:
            self._log_out.foreach(self.log.info)
        if self.tmux_nvim:
            self._cleanup_tmux()

    def _pre_start_neovim(self) -> None:
        pass

    def _post_start_neovim(self) -> None:
        rtp = fixture_path('config', 'rtp')
        self.vim.options.amend_l('runtimepath', rtp)

    def _pre_start(self) -> None:
        pass

    def _start_plugin(self) -> None:
        pass

    def _post_start(self) -> None:
        pass

    def _pvar_becomes(self, name, value) -> None:
        return self._wait_for(lambda: self.vim.vars.p(name).contains(value))

    def _pvar_becomes_map(self, name, value, f) -> None:
        return self._wait_for(
            lambda: self.vim.vars.p(name).map(f).contains(value))

    @property
    def _log_out(self) -> List[str]:
        return List.wrap(self.logfile.read_text().splitlines())

    def _json_cmd(self, cmd: str, data: dict) -> str:
        j = json.dumps(data).replace('"', '\\"')
        return '{} {}'.format(cmd, j)

    def json_cmd(self, cmd: str, **data: str) -> Either[Exception, str]:
        return self.vim.cmd(self._json_cmd(cmd, data))

    def json_cmd_sync(self, cmd: str, **data: str) -> Either[Exception, str]:
        return self.vim.cmd_sync(self._json_cmd(cmd, data))

    @property
    def content(self) -> List[str]:
        return self.vim.buffer.content


def main_looped(fun):
    @wraps(fun)
    def wrapper(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        asyncio.get_child_watcher().attach_loop(loop)
        done = asyncio.Future(loop=loop)
        exc = None
        ret = None
        def runner():
            nonlocal ret
            local_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(local_loop)
            try:
                ret = fun(self)
            except Exception as e:
                nonlocal exc
                exc = e
            finally:
                loop.call_soon_threadsafe(lambda: done.set_result(True))
                local_loop.close()
        Thread(target=runner).start()
        loop.run_until_complete(done)
        loop.close()
        if exc is not None:
            raise exc
        return ret
    return wrapper


@contextmanager
def _nop_main_loop(self):
    yield


def _mock_async(self, f):
    ret = f(self)
    return ret


def _mock_proxy(self):
    return self


class ExternalIntegrationSpec(VimIntegrationSpec):

    def __init__(self) -> None:
        VimIntegrationSpec.__init__(self)
        self._report_expensive = False

    def _pre_start_neovim(self):
        super()._pre_start_neovim()
        ribosome.in_vim = False
        NvimFacade.async = _mock_async
        NvimFacade.main_event_loop = _nop_main_loop
        NvimFacade.proxy = property(_mock_proxy)
        NvimFacade.clean = lambda self: True
        AsyncVimProxy.allow_async_relay = False

    def _post_start_neovim(self):
        cls = self.plugin_class.get_or_raise
        self.plugin = cls(self.neovim)

    def _await(self):
        if self.root is not None:
            self.root.await_state()

    @property
    def root(self):
        return self.plugin.state

    def teardown(self):
        VimIntegrationSpec.teardown(self)
        if self.root is not None:
            self.root.stop(shutdown=False)
            if self._debug and self._report_expensive:
                self._report()

    def _report(self):
        self.root.sub.cat(self.root) % __.report()


class PluginIntegrationSpec(VimIntegrationSpec):

    def _post_start_neovim(self):
        super()._post_start_neovim()
        self._setup_handlers()

    def _setup_handlers(self):
        rp_path = self.rplugin_path.get_or_raise
        rp_handlers = self.handlers.get_or_raise
        self.vim.call(
            'remote#host#RegisterPlugin',
            'python3',
            str(rp_path),
            rp_handlers,
        )

    @property
    def plugin_class(self) -> Either[str, type]:
        name = self.__class__.__name__
        e = 'property {}.plugin_class must return amino.Right(YourPluginClass)'
        return Left(e.format(name))

    @property
    def rplugin_path(self) -> Either[str, Path]:
        return self.plugin_class / self._auto_rplugin

    @property
    def handlers(self):
        return self.plugin_class / self._auto_handlers

    def _auto_handlers(self, cls):
        import inspect
        return List.wrap(inspect.getmembers(cls)).flat_map2(self._auto_handler)

    def _auto_handler(self, method_name, fun):
        fix = lambda v: int(v) if isinstance(v, bool) else v
        m = Maybe(getattr(fun, '_nvim_rpc_spec', None))
        return m / Map / __.valmap(fix)

    def _auto_rplugin(self, cls):
        mod = cls.__module__
        name = cls.__name__
        rp_path = temp_file('ribosome', 'spec', 'plugin.py')
        rp_path.write_text(rplugin_template.format(plugin_module=mod, plugin_class=name))
        return rp_path

__all__ = ('VimIntegrationSpec', 'PluginIntegrationSpec')
